import json

import httpx
import pytest

from fbnconfig import datatype as dt
from fbnconfig import property as prop
from fbnconfig import reference_list as rl

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeReferenceListResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with a string list
        sut = rl.ReferenceListResource(
            id="ref1",
            scope="scope-a",
            code="code-a",
            name="Test Reference List",
            description="A test reference list",
            tags=["test", "example"],
            reference_list=rl.StringList(values=["value1", "value2", "value3"])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-a"
        assert state["code"] == "code-a"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-a", "code": "code-a"}
        assert request_body["name"] == "Test Reference List"
        assert request_body["description"] == "A test reference list"
        assert request_body["tags"] == ["test", "example"]
        assert request_body["referenceList"]["values"] == ["value1", "value2", "value3"]
        assert request_body["referenceList"]["referenceListType"] == "StringList"

    def test_create_with_portfolio_group_id_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with a portfolio group id list
        sut = rl.ReferenceListResource(
            id="ref2",
            scope="scope-b",
            code="code-b",
            name="Portfolio Group Reference List",
            reference_list=rl.PortfolioGroupIdList(values=[
                rl.ResourceId(scope="group-scope", code="group-1"),
                rl.ResourceId(scope="group-scope", code="group-2")
            ])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-b"
        assert state["code"] == "code-b"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-b", "code": "code-b"}
        assert request_body["name"] == "Portfolio Group Reference List"
        assert request_body["referenceList"]["values"] == [
            {"scope": "group-scope", "code": "group-1"},
            {"scope": "group-scope", "code": "group-2"}
        ]
        assert request_body["referenceList"]["referenceListType"] == "PortfolioGroupIdList"

    def test_create_with_portfolio_id_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with a portfolio id list
        sut = rl.ReferenceListResource(
            id="ref3",
            scope="scope-c",
            code="code-c",
            name="Portfolio Reference List",
            reference_list=rl.PortfolioIdList(values=[
                rl.ResourceId(scope="portfolio-scope", code="portfolio-1"),
                rl.ResourceId(scope="portfolio-scope", code="portfolio-2")
            ])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-c"
        assert state["code"] == "code-c"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-c", "code": "code-c"}
        assert request_body["name"] == "Portfolio Reference List"
        assert request_body["referenceList"]["values"] == [
            {"scope": "portfolio-scope", "code": "portfolio-1"},
            {"scope": "portfolio-scope", "code": "portfolio-2"}
        ]
        assert request_body["referenceList"]["referenceListType"] == "PortfolioIdList"

    def test_create_with_decimal_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with a decimal list
        sut = rl.ReferenceListResource(
            id="ref4",
            scope="scope-d",
            code="code-d",
            name="Decimal Reference List",
            reference_list=rl.DecimalList(values=[1.5, 2.75, 3.0, 4.25])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-d"
        assert state["code"] == "code-d"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-d", "code": "code-d"}
        assert request_body["name"] == "Decimal Reference List"
        assert request_body["referenceList"]["values"] == [1.5, 2.75, 3.0, 4.25]
        assert request_body["referenceList"]["referenceListType"] == "DecimalList"

    def test_create_with_fund_id_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with a fund id list
        sut = rl.ReferenceListResource(
            id="ref5",
            scope="scope-e",
            code="code-e",
            name="Fund Reference List",
            reference_list=rl.FundIdList(values=[
                rl.ResourceId(scope="fund-scope", code="fund-1"),
                rl.ResourceId(scope="fund-scope", code="fund-2")
            ])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-e"
        assert state["code"] == "code-e"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-e", "code": "code-e"}
        assert request_body["name"] == "Fund Reference List"
        assert request_body["referenceList"]["values"] == [
            {"scope": "fund-scope", "code": "fund-1"},
            {"scope": "fund-scope", "code": "fund-2"}
        ]
        assert request_body["referenceList"]["referenceListType"] == "FundIdList"

    def test_create_with_address_key_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with an address key list
        sut = rl.ReferenceListResource(
            id="ref6",
            scope="scope-f",
            code="code-f",
            name="Address Key Reference List",
            reference_list=rl.AddressKeyList(values=["address-key-1", "address-key-2", "address-key-3"])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-f"
        assert state["code"] == "code-f"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-f", "code": "code-f"}
        assert request_body["name"] == "Address Key Reference List"
        assert request_body["referenceList"]["values"] == [
            "address-key-1", "address-key-2", "address-key-3"
        ]
        assert request_body["referenceList"]["referenceListType"] == "AddressKeyList"

    def test_create_with_instrument_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a desired reference list with an instrument list
        sut = rl.ReferenceListResource(
            id="ref7",
            scope="scope-g",
            code="code-g",
            name="Instrument Reference List",
            reference_list=rl.InstrumentList(values=["LUID_12345", "LUID_67890", "LUID_11111"])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-g"
        assert state["code"] == "code-g"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-g", "code": "code-g"}
        assert request_body["name"] == "Instrument Reference List"
        assert request_body["referenceList"]["values"] == ["LUID_12345", "LUID_67890", "LUID_11111"]
        assert request_body["referenceList"]["referenceListType"] == "InstrumentList"

    def test_create_with_property_list(self, respx_mock):
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given property definitions for the property list
        prop_def = prop.DefinitionResource(
            id="test-prop",
            domain=prop.Domain.Portfolio,
            scope="test-scope",
            code="test-property",
            display_name="Test Property",
            data_type_id=dt.DataTypeRef(id="string-dt", scope="system", code="string"),
            property_description="A test property for reference lists"
        )
        # given a desired reference list with a property list
        sut = rl.ReferenceListResource(
            id="ref8",
            scope="scope-h",
            code="code-h",
            name="Property Reference List",
            reference_list=rl.PropertyList(values=[
                rl.PropertyListItem(
                    property_definition=prop_def,
                    label_value="test-label-1",
                    metric_value=None,
                    label_set_value=None
                ),
                rl.PropertyListItem(
                    property_definition=prop_def,
                    label_value="test-label-2",
                    metric_value=None,
                    label_set_value=None
                )
            ])
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned with content hash
        assert state["scope"] == "scope-h"
        assert state["code"] == "code-h"
        assert "content_hash" in state
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-h", "code": "code-h"}
        assert request_body["name"] == "Property Reference List"
        assert len(request_body["referenceList"]["values"]) == 2
        # Check the first property list item
        first_item = request_body["referenceList"]["values"][0]
        assert first_item["key"] == "Portfolio/test-scope/test-property"
        assert first_item["labelValue"] == "test-label-1"
        assert "metricValue" not in first_item
        assert "labelSetValue" not in first_item
        # Check the second property list item
        second_item = request_body["referenceList"]["values"][1]
        assert second_item["key"] == "Portfolio/test-scope/test-property"
        assert second_item["labelValue"] == "test-label-2"
        assert "metricValue" not in second_item
        assert "labelSetValue" not in second_item

    def test_deps_with_string_list(self):
        # given a reference list with a string list
        sut = rl.ReferenceListResource(
            id="ref1",
            scope="scope-a",
            code="code-a",
            name="Test Reference List",
            reference_list=rl.StringList(values=["value1", "value2", "value3"])
        )
        # when we get its dependencies
        deps = sut.deps()
        # then it has no dependencies
        assert deps == []

    def test_deps_with_property_list(self):
        # given property definitions for the property list
        prop_def1 = prop.DefinitionResource(
            id="test-prop1",
            domain=prop.Domain.Portfolio,
            scope="test-scope",
            code="test-property-1",
            display_name="Test Property 1",
            data_type_id=dt.DataTypeRef(id="string-dt", scope="system", code="string"),
            property_description="First test property"
        )
        prop_def2 = prop.DefinitionRef(
            id="test-prop2",
            domain=prop.Domain.Portfolio,
            scope="test-scope",
            code="test-property-2",
        )
        # given a reference list with a property list using multiple property definitions
        sut = rl.ReferenceListResource(
            id="ref8",
            scope="scope-h",
            code="code-h",
            name="Property Reference List",
            reference_list=rl.PropertyList(values=[
                rl.PropertyListItem(
                    property_definition=prop_def1,
                    label_value="test-label-1",
                    metric_value=None,
                    label_set_value=None
                ),
                rl.PropertyListItem(
                    property_definition=prop_def2,
                    label_value="test-label-2",
                    metric_value=None,
                    label_set_value=None
                )
            ])
        )
        # when we get its dependencies
        deps = sut.deps()
        # then it should return both property definitions
        assert len(deps) == 2
        assert prop_def1 in deps
        assert prop_def2 in deps

    def test_deps_with_property_list_duplicate_properties(self):
        # given a property definition that will be used multiple times
        prop_def = prop.DefinitionResource(
            id="test-prop",
            domain=prop.Domain.Portfolio,
            scope="test-scope",
            code="test-property",
            display_name="Test Property",
            data_type_id=dt.DataTypeRef(id="string-dt", scope="system", code="string"),
            property_description="A test property"
        )
        # given a reference list with a property list using the same property definition multiple times
        sut = rl.ReferenceListResource(
            id="ref8",
            scope="scope-h",
            code="code-h",
            name="Property Reference List",
            reference_list=rl.PropertyList(values=[
                rl.PropertyListItem(
                    property_definition=prop_def,
                    label_value="test-label-1",
                    metric_value=None,
                    label_set_value=None
                ),
                rl.PropertyListItem(
                    property_definition=prop_def,
                    label_value="test-label-2",
                    metric_value=None,
                    label_set_value=None
                ),
                rl.PropertyListItem(
                    property_definition=prop_def,
                    label_value="test-label-3",
                    metric_value=None,
                    label_set_value=None
                )
            ])
        )
        # when we get its dependencies
        deps = sut.deps()
        # then it should return the property definition only once (deduplicated)
        assert len(deps) == 1
        assert deps[0] == prop_def

    def test_update_with_no_changes(self, respx_mock):
        from types import SimpleNamespace
        # given a reference list with a string list
        sut = rl.ReferenceListResource(
            id="ref1",
            scope="scope-a",
            code="code-a",
            name="Test Reference List",
            description="A test reference list",
            tags=["test", "example"],
            reference_list=rl.StringList(values=["value1", "value2", "value3"])
        )
        # and an old state with the same content hash
        desired = sut.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        from hashlib import sha256
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        old_state = SimpleNamespace(scope="scope-a", code="code-a", content_hash=content_hash)
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None (no changes needed)
        assert state is None
        # and no HTTP requests were made
        assert len(respx_mock.calls) == 0

    def test_update_with_changes(self, respx_mock):
        from types import SimpleNamespace
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a reference list with a string list
        sut = rl.ReferenceListResource(
            id="ref1",
            scope="scope-a",
            code="code-a",
            name="Test Reference List",
            description="A test reference list",
            tags=["test", "example"],
            reference_list=rl.StringList(values=["value1", "value2", "value3"])
        )
        # and an old state with a different content hash
        old_state = SimpleNamespace(scope="scope-a", code="code-a", content_hash="different_hash")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is returned with new content hash
        assert state is not None
        assert state["scope"] == "scope-a"
        assert state["code"] == "code-a"
        assert "content_hash" in state
        assert state["content_hash"] != "different_hash"
        # and a POST request was made to update the reference list
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/referencelists"
        request_body = json.loads(request.content)
        assert request_body["id"] == {"scope": "scope-a", "code": "code-a"}
        assert request_body["name"] == "Test Reference List"

    def test_update_with_scope_code_change(self, respx_mock):
        from types import SimpleNamespace
        respx_mock.delete("/api/api/referencelists/old-scope/old-code").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/referencelists").mock(return_value=httpx.Response(200, json={}))
        # given a reference list with new scope/code
        sut = rl.ReferenceListResource(
            id="ref1",
            scope="new-scope",
            code="new-code",
            name="Test Reference List",
            reference_list=rl.StringList(values=["value1", "value2", "value3"])
        )
        # and an old state with different scope/code
        old_state = SimpleNamespace(scope="old-scope", code="old-code", content_hash="old_hash")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the old one is deleted and a new one is created
        assert state is not None
        assert state["scope"] == "new-scope"
        assert state["code"] == "new-code"
        assert "content_hash" in state
        # and both DELETE and POST requests were made
        assert len(respx_mock.calls) == 2
        delete_request = respx_mock.calls[0].request
        assert delete_request.method == "DELETE"
        assert delete_request.url.path == "/api/api/referencelists/old-scope/old-code"
        create_request = respx_mock.calls[1].request
        assert create_request.method == "POST"
        assert create_request.url.path == "/api/api/referencelists"

    def test_delete(self, respx_mock):
        from types import SimpleNamespace
        respx_mock.delete("/api/api/referencelists/scope-a/code-a").mock(
            return_value=httpx.Response(200, json={})
        )
        # given a resource that exists in the remote
        old_state = SimpleNamespace(scope="scope-a", code="code-a")
        # when we delete it
        rl.ReferenceListResource.delete(self.client, old_state)
        # then a DELETE request was made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/api/api/referencelists/scope-a/code-a"

    def test_create_property_list_with_both_metric_and_label_values_throws(self):
        # given property definitions for the property list
        prop_def = prop.DefinitionResource(
            id="test-prop",
            domain=prop.Domain.Portfolio,
            scope="test-scope",
            code="test-property",
            display_name="Test Property",
            data_type_id=dt.DataTypeRef(id="string-dt", scope="system", code="string"),
            property_description="A test property for reference lists"
        )
        # when we try to create a PropertyListItem with both metric_value and label_value set
        with pytest.raises(KeyError, match="Cannot set label_value and metric_value"):
            rl.PropertyListItem(
                property_definition=prop_def,
                label_value="test-label",
                metric_value=rl.MetricValue(value=123.45, unit="USD"),
                label_set_value=None
            )



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
import tests.integration.property as property
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("property")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    # create first
    fbnconfig.deploy(property.configure(setup_deployment), lusid_env, token)
    fbnconfig.deploy(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    with pytest.raises(HTTPStatusError) as error:
        client.get(f"/api/api/propertydefinitions/Holding/{setup_deployment.name}/more_derived")
    assert error.value.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deploy(property.configure(setup_deployment), lusid_env, token)
    search = client.get(f"/api/api/propertydefinitions/Holding/{setup_deployment.name}/more_derived")
    assert search.status_code == 200


def test_update(setup_deployment):
    fbnconfig.deploy(property.configure(setup_deployment), lusid_env, token)
    fbnconfig.deploy(property.configure(setup_deployment), lusid_env, token)
    search = client.get(f"/api/api/propertydefinitions/Holding/{setup_deployment.name}/more_derived")
    assert search.status_code == 200



from __future__ import annotations

import copy
from enum import StrEnum
from typing import Any, Dict, Optional

import httpx
from pydantic import BaseModel, Field

from .resource_abc import Ref, Resource, register_resource


@register_resource()
class IdentityRoleRef(BaseModel, Ref):
    """Used to reference an existing identity role for role assignment
    Example
    -------
    >>> from fbnconfig.identity import IdentityRoleRef
    >>> admin = IdentityRoleRef(id="admin", name="lusid-administrator")
    """

    role_id: str | None = Field(None, exclude=True, init=False)
    name: str
    id: str = Field(exclude=True)

    def attach(self, client):
        search = client.request("get", "/identity/api/roles")
        try:
            match = next(role for role in search.json() if role["name"] == self.name)
        except StopIteration:
            raise (RuntimeError("No matching identity role with name", self.name))
        self.role_id = match["id"]


@register_resource()
class IdentityRoleResource(BaseModel, Resource):
    """Manage a role

    Example
    -------
    >>> from fbnconfig.identity import IdentityRoleResource
    >>> operator =
    >>> IdentityRoleResource(id="export-role", description="scheduler-operators", name="export-admin")
    """

    id: str = Field(exclude=True)
    name: str
    description: str
    scope: str | None = Field(None, exclude=True, init=False)
    code: str | None = Field(None, exclude=True, init=False)
    role_id: str | None = Field(None, exclude=True, init=False)
    remote: Dict[str, Any] = Field(None, exclude=True, init=False)

    def read(self, client, old_state):
        get = client.request("get", f"/identity/api/roles/{old_state.roleId}")
        remote = get.json()
        remote.pop("id")
        remote.pop("name")
        remote.pop("roleId")
        remote.pop("source")
        remote.pop("samlName")
        self.remote = remote
        self.role_id = old_state.roleId

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True)
        res = client.request("POST", "/identity/api/roles", json=desired)
        result = res.json()
        self.scope = result["roleId"]["scope"]
        self.code = result["roleId"]["code"]
        self.role_id = res.json()["id"]
        return {
            "id": self.id,
            "code": self.code,
            "scope": self.scope,
            "name": self.name,
            "roleId": self.role_id,
        }

    def update(self, client: httpx.Client, old_state):
        if self.name != old_state.name:
            raise (RuntimeError("Cannot change the name on a role"))
        self.read(client, old_state)
        remote = copy.deepcopy(self.remote)
        self.role_id = old_state.roleId
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"name"})

        if desired == remote:
            return None
        client.request("put", f"/identity/api/roles/{old_state.roleId}", json=desired)

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/identity/api/roles/{old_state.roleId}")

    def deps(self):
        return []


class UserType(StrEnum):
    """User type used with a UserResource"""

    PERSONAL = "Personal"
    SERVICE = "Service"


@register_resource()
class UserRef(BaseModel, Ref):
    """Reference an existing user

    Example
    -------
    >>> from fbnconfig.identity import UserRef
    >>> user =  UserRef(id="export-user", login="exports@company.com")
    """

    id: str = Field(exclude=True)
    login: str
    user_id: str | None = Field(None, exclude=True, init=False)

    def attach(self, client):
        user_list = client.request("get", "/identity/api/users")
        match = next((user for user in user_list.json() if user["login"] == self.login), None)
        if match is None:
            raise RuntimeError(f"No user with login {self.login} exists in the remote")
        self.user_id = match["id"]


@register_resource()
class UserResource(BaseModel, Resource):
    """Define a user"""

    id: str = Field(exclude=True)
    first_name: str
    last_name: str
    email_address: str
    second_email_address: Optional[str] = None
    login: str
    type: UserType
    user_id: str | None = Field(None, exclude=True, init=False)
    remote: Dict[str, Any] = Field({}, exclude=True, init=False)

    def read(self, client, old_state):
        get = client.request("get", f"/identity/api/users/{old_state.userId}")
        remote = get.json()
        remote.pop("status", None)
        remote.pop("external", None)
        remote.pop("id", None)
        remote.pop("roles", None)
        remote.pop("links", None)
        self.remote = remote
        self.user_id = old_state.userId

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        res = client.request("POST", "/identity/api/users", json=desired)
        result = res.json()
        self.user_id = result["id"]
        return {"id": self.id, "userId": self.user_id}

    def update(self, client: httpx.Client, old_state):
        self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        if desired == self.remote:
            return None
        if self.login != self.remote["login"]:
            raise RuntimeError("Cannot change the login on an existing user")
        client.request("put", f"/identity/api/users/{old_state.userId}", json=desired)
        return {"id": self.id, "userId": self.user_id}

    @staticmethod
    def delete(client, old_state):
        client.request("delete", f"/identity/api/users/{old_state.userId}")

    def deps(self):
        return []


@register_resource()
class RoleAssignment(BaseModel, Resource):
    """Assign a role to a user"""

    id: str = Field(exclude=True)
    user: UserResource | UserRef
    role: IdentityRoleResource | IdentityRoleRef
    role_id: str | None = Field(None, exclude=True, init=False)
    user_id: str | None = Field(None, exclude=True, init=False)
    remote: Dict[str, Any] = Field(None, exclude=True, init=False)

    def read(self, client, old_state):
        self.role_id = old_state.roleId
        self.user_id = old_state.userId

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        role_id = self.role.role_id
        user_id = self.user.user_id
        client.request("put", f"/identity/api/roles/{role_id}/users/{user_id}")
        return {"id": self.id, "userId": user_id, "roleId": role_id}

    @staticmethod
    def delete(client: httpx.Client, old_state) -> None:
        role_id = old_state.roleId
        user_id = old_state.userId
        client.request("delete", f"/identity/api/roles/{role_id}/users/{user_id}")

    def update(self, client, old_state) -> Dict[str, Any] | None:
        if old_state.userId == self.user.user_id and old_state.roleId == self.role.role_id:
            return None
        try:
            client.request("delete", f"/identity/api/roles/{old_state.roleId}/users/{old_state.userId}")
        except httpx.HTTPStatusError as e:
            # ignore 404 because if the old role or user are gone this won't exist
            if e.response.status_code != httpx.codes.NOT_FOUND:
                raise
            pass
        client.request("put", f"/identity/api/roles/{self.role.role_id}/users/{self.user.user_id}")
        return {"id": self.id, "userId": self.user.user_id, "roleId": self.role.role_id}

    def deps(self):
        return [self.user, self.role]



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import sequence

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSequenceResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/api/api/sequences/scope-a").mock(return_value=httpx.Response(200, json={}))
        # given a desired sequence where we default the startValue
        sut = sequence.SequenceResource(
            id="xyz",
            scope="scope-a",
            code="code-a",
            increment=2,
            min_value=1,
            max_value=100,
            pattern="SQP-{{seqValue}}",
        )
        # when we create it
        state = sut.create(self.client)
        # then the state is returned
        assert state == {"scope": "scope-a", "code": "code-a"}
        # and a create request was sent without the startValue
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/sequences/scope-a"
        assert json.loads(request.content) == {
            "scope": "scope-a",
            "code": "code-a",
            "increment": 2,
            "minValue": 1,
            "maxValue": 100,
            "pattern": "SQP-{{seqValue}}",
        }

    def test_update_with_no_changes(self, respx_mock):
        # given an existing sequence
        respx_mock.get("/api/api/sequences/scope-a/code-a").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"scope": "scope-a", "code": "code-a"},
                    "increment": 2,
                    "startValue": 1,
                    "minValue": 1,
                    "maxValue": 100,
                    "pattern": "SQP-{{seqValue}}",
                },
            )
        )
        # and a desired with increment but everything else defaulted
        sut = sequence.SequenceResource(id="seq2", scope="scope-a", code="code-a", increment=2)
        old_state = SimpleNamespace(scope="scope-a", code="code-a")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and a read was made but no PUT

    def test_update_with_changes_throws(self, respx_mock):
        # given an existing sequence
        respx_mock.get("/api/api/sequences/scope-a/code-a").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"scope": "scope-b", "code": "code-b"},
                    "scope": "scope-a",
                    "code": "code-a",
                    "increment": 2,
                    "startValue": 1,
                    "minValue": 1,
                    "maxValue": 100,
                    "pattern": "SQP-{{seqValue}}",
                },
            )
        )
        # and a desired with a different increment
        sut = sequence.SequenceResource(id="seq2", scope="scope-a", code="code-a", increment=200)
        old_state = SimpleNamespace(scope="scope-a", code="code-a")
        # when we update it throws
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_delete_throws(self):
        # given a resource that exists in the remnte
        old_state = SimpleNamespace(scope="scope-b", code="code-b")
        # when we delete it throws
        with pytest.raises(RuntimeError):
            sequence.SequenceResource.delete(self.client, old_state)

    def test_deps(self):
        sut = sequence.SequenceResource(id="xyz", scope="scope-b", code="code-b")
        # it's deps are empty
        assert sut.deps() == []



from abc import ABC, abstractmethod
from typing import Any, Dict

from polyfactory.factories.pydantic_factory import ModelFactory
from pydantic import AliasGenerator, BaseModel, ConfigDict, Field, alias_generators, computed_field

alias_generator = ConfigDict(
    alias_generator=AliasGenerator(
        serialization_alias=lambda field_name: alias_generators.to_camel(field_name)
    )
)


class BaseResource(ABC):
    model_config = alias_generator

    @abstractmethod
    def read(self) -> Any:
        pass


class SomeOtherObject(BaseModel):
    model_config = alias_generator
    some_other_field: str
    some: str


#
# when mapping multiple flat fields into a structure
# for the request, mark the fields the user sets
# as exclude=True and add a computed field
# to convert to the serialized representation
#
class ComputedResource(BaseResource, BaseModel):
    id: str = Field(exclude=True)
    scope: str = Field(init=True, exclude=True)
    code: str = Field(init=True, exclude=True)
    test_field: str
    other_field: SomeOtherObject

    @computed_field(alias="Id")
    def set_id(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self):
        return 3


def test_dump():
    c = ComputedResource(
        id="id",
        scope="scope1",
        code="code1",
        test_field="something",
        other_field=SomeOtherObject(some_other_field="myvalue", some="testings"),
    )

    desired = c.model_dump(mode="json", exclude_none=True, by_alias=True)
    assert desired == {
        "Id": {"scope": "scope1", "code": "code1"},
        "testField": "something",
        "otherField": {"someOtherField": "myvalue", "some": "testings"},
    }


def test_computed_factory() -> None:
    class ComputedRFactory(ModelFactory[ComputedResource]): ...

    computed_instance = ComputedRFactory.build()
    assert isinstance(computed_instance, ComputedResource)
    assert isinstance(computed_instance.id, str)
    assert isinstance(computed_instance.scope, str)
    assert isinstance(computed_instance.code, str)
    assert isinstance(computed_instance.test_field, str)
    assert isinstance(computed_instance.other_field, SomeOtherObject)
